/*
Copyright (C) 2006 Paul Thomas
This file may be copied and redistributed freely, without advance permission,
provided that this Copyright statement is reproduced with each copy.

LIMITATION OF WARRANTY
NOTHING IN THIS AGREEMENT WILL BE CONSTRUED AS A REPRESENTATION MADE OR
WARRANTY GIVEN BY PAUL THOMAS OR ANY THIRD PARTY THAT THE USE OF
DATA PROVIDED HEREUNDER WILL NOT INFRINGE ANY PATENT, COPYRIGHT, TRADEMARK
OR OTHER RIGHTS OF ANY THIRD PARTY. DATA IS PROVIDED "AS IS" WITHOUT
WARRANTY OF ANY KIND WHATSOEVER, EXPRESS OR IMPLIED, INCLUDING IMPLIED
WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. PAUL
THOMAS MAKES NO WARRANTY THAT ITS DATA DOES NOT CONTAIN ERRORS.
*/


/* This program takes a multiple alignment in FASTA or PIR format and 
calculates sequence weights according to:
Henikoff & Henikoff, JMB 243: 574 (1994)
output is designed for direct input into SAM
*/
#include "gen_incl.h"
#include "fasta_seq.h"

#define OPTSTRING "m:HNe:hc:i:"
extern char *optarg;

#define MAXNSEQS 10000

FILE *err_file; /* global variable for controlling error output */


void main(int argc, char *argv[]) {
   int i, j, option, end;
   char c;
   int nseq; /* number of sequences in multiple alignment */
   fasta_seq *seq[MAXNSEQS]; /* sequences */
   double wt[MAXNSEQS]; /* weight of each sequence */
   double total_wt;
   int count[27]; /* amino acid counts */
   int total; /* total different amino acids at given position */
   double cst=0.0;
   FILE *in_file=NULL;
   int lseq;
   int high;
   int henikoff = 0; /* print Henikoff weights */
   int normSeqs = 0; /* print Henikoff weights, weighted to number of seqs*/
   double maxcst=1000000.; /* maximum number of independent counts */
   double cons=0.; /* conservation */
   int UNALIGNED='A'+26; /* dummy amino acid for an unaligned position */
   int nalignedseq; /* number of sequences aligned at a given position */
   void usage();
   int lalign=0;

   /* get options */
   while((option=getopt(argc,argv,OPTSTRING)) !=EOF) {
     switch(option) {
     case 'e':	/* Error file */
       if (*optarg == '-') err_file = stdin;
       else  err_file = fopen(optarg,"a"); 
       break;
     case 'i':	/* Input file */
       in_file = fopen(optarg,"r"); 
       break;
     case 'h':
       usage(); exit(0);
       break;
     case 'N':
       normSeqs = 1;
       break;
     case 'H':
       henikoff = 1;
       break;
     case 'm':
       if(sscanf(optarg,"%lf",&maxcst)!=1)
	 fprintf(stderr,"Warning: -m (%s) must be a number.\n Using default.\n", optarg);
       break;
     case 'c':
       if(sscanf(optarg,"%lf",&cst)!=1)
	 fprintf(stderr,"Warning: -c (%s) must be a number.\n Using default.\n", optarg);
       break;
     default :
       fprintf(stderr,"Error: do not understand -%c option\n");
       usage(); exit(1);
       break;
     }
   }

   if(in_file==NULL) {
      fprintf(stderr,"Error-- no valid input file\n");
      usage();
      exit(1);
   }

   nseq = 0;
   while(read_fasta_seq(in_file, &seq[nseq], ALIGN) != EOF) {
      /* convert dashes to an UNALIGNED (dummy) amino acid type */
      for(j=0;j<seq[nseq]->seqlen;++j) {
	 if(seq[nseq]->seq[j]=='-') 
	    seq[nseq]->seq[j]=UNALIGNED;
      }
/*
      printf(">%s\n%s\n", seq[nseq]->comment, seq[nseq]->seq);
*/
      nseq++;
      if(end==EOF) break;
   }

   for(i=0;i<nseq;++i) wt[i] = 0.;

   /* check to make sure all sequences are the same length */
   lseq = seq[0]->seqlen;
   for(i=1;i<nseq;++i) {
      if(seq[i]->seqlen!=lseq) { 
	 fprintf(stderr,"Error: sequence lengths differ!\n");
	 exit(1);
      }
   }

   /* calculate weights for each position, then sum */
   for(j=0;j<lseq;++j) {
      for(i=0;i<27;++i) count[i]=0;
      nalignedseq=0;
      for(i=0;i<nseq;++i) {
	 count[seq[i]->seq[j]-'A']++;
      }
      for(i=0,total=0;i<26;++i) if(count[i]>0) total++;
      for(i=0;i<nseq;++i) {
	 if(seq[i]->seq[j]!=UNALIGNED) {
	    wt[i] += 1. / (double)total / (double)count[seq[i]->seq[j]-'A'];
	    ++nalignedseq;
	 }
      }
      /* determine frequency of highest-frequency amino acid */
      for(i=0,high=0;i<26;++i) high = MAX(high, count[i]);
      if(nalignedseq>0) {
         cons += (double)high / (double)nalignedseq;
	 ++lalign;
      }
   }

   /* print number of independent counts */
   cons /= (double)lalign;
   if(cst <= 0.0) { /* if cst was not set on the command line */
      /* allow no more than maxcst independent counts */
      if((cst =  pow((double)nseq, 1. - cons)) > maxcst) cst = maxcst;
   }
   printf("number of independent counts = %.2lf, alignment length = %d\n", 
      cst, lalign);

   /* SAM needs the number of sequences and the number of subfamilies */
   printf("%d 1\n",nseq);

   /* normalize weights to 1 */
   for(i=0,total_wt=0.;i<nseq;++i) total_wt += wt[i];

   /* print weights */
   for(i=0;i<nseq;++i) {
     if (henikoff) {
       printf("%s %.3f\n", seq[i]->comment, wt[i]/total_wt);
     } else if (normSeqs) {
       printf("%s %.3f\n", seq[i]->comment, nseq * (wt[i]/total_wt));
     } else {
       printf("%s %.3f\n", seq[i]->comment, cst*wt[i]/total_wt);
     }
   }
}

void usage() {
   printf("\nseq_wt -i <input_file> [options]\n");
   printf("\nCalculates relative sequence weights according to Henikoff & Henikoff, JMB 1994\n");
   printf("The algorithm is modified to treat gaps as having zero weight.\n");
   printf("The input must be a multiple alignment in FASTA or PIR format.\n");
   printf("Output goes to standard output.\n");
   printf("\nOptions:\n");
   printf("\t-h \t\tThis help message.\n");
   printf("\t-e filename\tSend error messages to file <filename> (Default: stderr)\n");
   printf("\t-c <float>\tScaling factor for normalized weights\n\t\t\t(Default = number of independent observations)\n");
   printf("\t-m <float>\tMaximum scaling factor for normalized weights\n\t\t\t(Default = number of independent observations)\n");
}
