#! /usr/local/bin/perl

# Copyright (C) 2006 Paul Thomas.
# This file may be copied and redistributed freely, without advance permission,
# provided that this Copyright statement is reproduced with each copy.

# LIMITATION OF WARRANTY
# NOTHING IN THIS AGREEMENT WILL BE CONSTRUED AS A REPRESENTATION MADE OR
# WARRANTY GIVEN BY PAUL THOMAS OR ANY THIRD PARTY THAT THE USE OF
# DATA PROVIDED HEREUNDER WILL NOT INFRINGE ANY PATENT, COPYRIGHT, TRADEMARK
# OR OTHER RIGHTS OF ANY THIRD PARTY. DATA IS PROVIDED "AS IS" WITHOUT
# WARRANTY OF ANY KIND WHATSOEVER, EXPRESS OR IMPLIED, INCLUDING IMPLIED
# WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. PAUL
# THOMAS MAKES NO WARRANTY THAT ITS DATA DOES NOT CONTAIN ERRORS.

#####
#####
##
## snp_analysis.pl - analyzes SNP results 
## providing HMM statics for a substitution based on the corresponding node.
##
## Anish Kejariwal - 9/9/01
#####
#####
    
$|=1;

# necessary libraries
use FamLibBuilder;
use FastaFile;
use Algs::Hmmer;
use strict;

# default values
my ($library,$classFile,$outFile,$snpFile,$fastaFile,$allSnps,$errFile,$verbose,$priorFile,$hmmsearchPath,$forceSfFlag);
my $type = "orig";
my $tmpdirPath = "/tmp";
my $ghmmType = "mag"; 
my $blosum62File = "$ENV{'BLASTHOME'}/matrix/aa/BLOSUM62";
my %validAAs = ('A' => 1, 'C' => 1, 'D' => 1, 'E' => 1, 'F' => 1, 'G' => 1, 'H' => 1, 'I' => 1, 'K' => 1, 'L' => 1, 'M' => 1, 'N' => 1, 'P' => 1, 'Q' => 1, 'R' => 1, 'S' => 1, 'T' => 1, 'V' => 1, 'W' => 1, 'Y' => 1,'*' => 1, );
my $prod = 1;
my $scoreCut = 1.0e-23;

# get command-line arguments
use Getopt::Long; 
$Getopt::Long::ignorecase=0; #will not ignore case
&GetOptions
(
"h" => sub {&usage()},   # -h for help
"l=s" => \$library,      # -l for (l)ibrary
"t=s" => \$type,         # -t for entry (t)ype
 "T=s" => \$tmpdirPath,  # -T for path to (T)mp directory
"c=s" => \$classFile,    # -c for (c)lass.ts file
"o=s" => \$outFile,      # -o for (o)utput file (redirect STDOUT)
"s=s" => \$snpFile,      # -s for (s)np file 
"f=s" => \$fastaFile,    # -f for (f)asta file of seqs to be analyzed
"P=i" => \$prod,         # -P for (P)roduction version of library
"a" => \$allSnps,        # -a for (a)ll snps (not just best prot)
"b=s" => \$blosum62File, # -b for (b)losum 62 matrix file
"S" => \$forceSfFlag,    # -S force snp analysis to be on (S)F and not use evol SF
"p=s" => \$priorFile,    # -p for (p)rior file
"H=s" => \$hmmsearchPath,  # -H for path to (H)mmsearch binary 
"e" => \$errFile,        # -e for (e)rror file (redirect STDERR)
"V" => \$verbose         # -v for (v)erbose (debug info to STDERR)
);

&usage("Please specify SNP file\n") unless (-s $snpFile);
&usage("Please specify library\n") unless (-d $library);
&usage("Please specify class.ts file\n") unless (-s $classFile);
&usage("Please specify entry type\n") unless ($type);
&usage("Please specify fasta file\n") unless ($fastaFile);
&usage("Please specify BLOSUM62 matrix file\n") unless (-s $blosum62File);
&usage("Please specify path to prior file\n") unless (-s $priorFile);
&usage("tmp directory $tmpdirPath cannot be found\n") unless (-d $tmpdirPath);

# do file redirects
if ($outFile) {
    open (STDOUT, ">$outFile") || &usage("Cannot open $outFile");
}
if ($errFile) {
    open (STDERR, ">$errFile") || &usage("Cannot open $errFile");
}

&pukeSetup() if ($verbose);

#######################################################
###  MAIN BEGINS HERE
#######################################################

#create FamLibBuilder object access files in library
my $flb;
if ($prod) {
    $flb = new FamLibBuilder($library,"prod");
} else {
    $flb = new FamLibBuilder($library);
}
&usage("Cannot find library $library.\n") unless ($flb->exists()); 

#read in BLOSUM62 matrix
my %blosumHash = &storeBlosum($blosum62File);

#read in fasta file of sequences to be analyzed, into a hash for easy access
print STDERR "Reading in fasta file\n" if ($verbose);
my $idToSeq = &readFastaFile($fastaFile);

#read in class.ts file, - get back best ghmm/shmm (and scores) for each seq
#structure:  $hmmScore{$seqId}{'ghmm'|'shmm'}{'id'|score'}
#also, check that everything in class.ts file exists in %idToSeq
print STDERR "Reading in class.ts file\n" if ($verbose);
my $hmmScores = &readClassFile($classFile,$idToSeq);

#for every snp, want just one protein in the results file - best=best ghmm scr
my $bestProtForSnp = {};
if (! $allSnps) {
  print STDERR "Calculating best prot for snp\n" if ($verbose);
  $bestProtForSnp = &getBestProtForSnp($hmmScores,$snpFile);
}

#print header
my @headerArray;
my @headerArray=("snpId","seqId","subPSEC","Pdeleterious","wtAA","aaPos","subAA","HMM ID","HMM name","HMM Score","hmmPos","msaPos","sfConsAA","Pwt","Psub","evolSFs","evolFam","NIC","blosum","message");
print join("\t",@headerArray),"\n";

#start reading in SNP file
open(SNP,"$snpFile");
my $prevLine;
while(my $line = <SNP>) {
  chomp $line;
  next unless ($line=~/\w/);
  $line=~s/[\cK]+//g;
  $line =~s/[\r]//g;
  $line =~ s/[\cM]+//g;
  
  my ($snp,$gi,$pos,$aaListString) = split(/\|/,$line);
   
  #only analyze best protein for snp
  #unless, used -a option to show all snp results
  if ( (! $allSnps) && ($gi ne $bestProtForSnp->{$snp})) {
    print STDERR "Skipping $snp\t$gi\t$pos\n" if ($verbose);
    next;
  } else {
    print STDERR "Working on $snp\t$gi\t$pos\n" if ($verbose);
  }
 
  $aaListString=~tr/[a-z]/[A-Z]/;
  
  my $output = {};
  $output->{"snpId"} = $snp;
  $output->{"seqId"} = $gi;
  $output->{"aaPos"} = $pos;
  my @aaList = split(/;/,$aaListString);

  #print out HMM info
  my $hmmId = $hmmScores->{$gi}{'id'};
  $output->{"HMM ID"} = $hmmId;
  my $hmmScore = $hmmScores->{$gi}{'score'};
  $output->{"HMM Score"} = $hmmScore;
  $output->{"HMM name"} = $hmmScores->{$gi}{'name'};

  #check that amino acids are valid
  if ( (! defined(@aaList)) || (scalar(@aaList) < 2)) {
    $output->{"message"} = "wild type and substituted amino acids required";
    print &createOutput($output,$aaList[0],$aaList[1]),"\n";
    next;
  }

  if (! &isValidAAs(\@aaList)) {
    $output->{"message"} = "invalid amino acid";
    print &createOutput($output,$aaList[0],$aaList[1]),"\n";
    next;
  }

  #get sequence
  my $seq = $idToSeq->{$gi};

  #if missing sequence cannot do anything
  unless ($seq) {
      $output->{"message"} = "Missing sequence";
    print &createOutput($output,$aaList[0],$aaList[1]),"\n";
    next;
  }    

  #check if SNP protein position is valid
  if (! &isSnpPositionValid($pos,$seq)) {
    $output->{"message"} = "SNP position not within protein";
    print &createOutput($output,$aaList[0],$aaList[1]),"\n";
    next; 
  }

  #figure out what is the wildtype aa
  my $wildTypeAA = &getWildTypeAA($gi,$pos,$seq);
  if ($wildTypeAA eq "*") {
    $output->{"message"} = "wild type amino acid is a stop codon";
    print &createOutput($output,$aaList[0],$aaList[1]),"\n";
    next;
  } elsif (! $wildTypeAA ) {
    $output->{"message"} = "SNP position does not exist in protein";
    print &createOutput($output,$aaList[0],$aaList[1]),"\n"; 
    next;
  }

  #one of the passed in amino acids must be the wild type AA
  my $foundWildType = 0;
  foreach my $aa (@aaList) {
    if ($wildTypeAA eq $aa) {
      $foundWildType = 1;
      last;
    }
  }
  if (! $foundWildType) {
    $output->{"message"} = "wild type amino acid is $wildTypeAA";
    print &createOutput($output,$aaList[0],$aaList[1]),"\n";  
    next;
  }
  
  my $reachedWt;
  foreach my $mutantAA (@aaList) {
    $output->{"wtAA"} = $wildTypeAA;
    $output->{"subAA"} = $mutantAA;
    if ($mutantAA eq $wildTypeAA) {
      if (! $reachedWt) {
	$reachedWt = 1;  #skip itself
	next;
      } else {
	$output->{"subPSEC"} = 0;
	$output->{"message"} = "silent mutation";
	print &createOutput($output),"\n";
	next;
      }
    }
    if ($mutantAA eq "*") {
      $output->{"message"} = "nonsense mutation";
      print &createOutput($output),"\n";
      next;
    }

    #get blosum score
    $output->{"blosum"} = $blosumHash{$wildTypeAA}{$mutantAA};
    
    #now need to figure out what HMM it hit, and get HMM file

    if (! $hmmId) {
      #weak score
      $output->{"message"} = "No PANTHER HMM hit";
      print &createOutput($output),"\n";
      next;
    }
    if ($hmmScore >= $scoreCut) {
	$output->{"message"} = "Weak PANTHER score";
	print &createOutput($output),"\n";
	next;
    } 

    my ($bookId,$sfId);
    if ($hmmId=~/^([^:]+):(SF\d+)$/) {
      $bookId = $1;
      $sfId = $2;
    } else {
      $bookId = $hmmId;
    }
    my $fle = $flb->getLibEntry($bookId,$type);
    my $hmmFile;
    if ($sfId) {
      $hmmFile = $fle->hmmerHmmFile($sfId);
    } else { 
      $hmmFile = $fle->hmmerHmmFile($ghmmType);
    }
    unless (-s $hmmFile ) {
      print STDERR "HMM ($hmmFile) missing\n";
      next;
    }
    
    #get HMM column
    #convert to MSA position
    #run Betty's thing and get probability
    my $hmm_col = &extractHmmerHmmPos($seq,$pos,$hmmFile);


    if (($hmm_col eq "insert state") || (! $hmm_col)) { 
      if ($hmm_col eq "insert state") { 
	$output->{"hmmPos"} = "insert state";
	$output->{"message"} = "SNP position in protein does not align to HMM"; 
      } else { 
	#generally only happens if position of snp is not in prot pos > length($seq) 
	$output->{"message"} = "substitution position incorrect"; 
      }
      print &createOutput($output),"\n";
      next;
    }
    $output->{"hmmPos"} = $hmm_col;
     
    #then parse HMMER HMM for MSA/SAM position
    #getMsaPos
    my $msaPos = &getMsaPos($hmm_col,$hmmFile);
    $output->{"msaPos"} = $msaPos;

    #what if msa pos not returned????

    #what if log min max is undef, or incorrect AA given to her program??

    #if SF or FAM get number of seqs from alignment file or tree?
    #GET IF FAM, SF, OR EVOL SF, WHICH SFS
    my $treeFile = $fle->treeFile();
    my $alignFile = $fle->alignFile();
    print STDERR "Missing tree file $treeFile\n" unless (-s $treeFile);
    print STDERR "Missing MSA file $alignFile\n" unless (-s $alignFile);
    my ($pWt,$pSub,$score,$Pdeleterious,$nic,$sfList,$evolFam,$sfConsAA,$newMsg) = &getEvolConsInfo($sfId,$treeFile,$alignFile,$msaPos,$wildTypeAA,$mutantAA);
    if ($nic == 0) {
        $output->{"hmmPos"} = "insert state";
        $output->{"message"} = "SNP position in protein does not align to HMM";
	print &createOutput($output),"\n";
	next;
    }


    $output->{"message"} .= $newMsg if ($newMsg);

    $output->{"evolFam"} = $evolFam;
    $output->{"Pwt"} = $pWt;
    $output->{"Psub"} = $pSub;
    $output->{"evolSFs"} = $sfList;
    $output->{"NIC"} = $nic;
    $output->{"subPSEC"} = $score;
    $output->{"Pdeleterious"} = $Pdeleterious;
    $output->{"sfConsAA"} = $sfConsAA;
    print &createOutput($output),"\n";
  }

}

close(SNP);


#######################################################
###  SUBROUTINES BEGIN HERE
#######################################################

#check that amino acids are valid
sub isValidAAs {
  my $aaList = shift;

  foreach my $aa (@$aaList) {
    if (! defined($validAAs{$aa})) {
      return 0;
    }
  }
  return 1;
}

sub getMsaPos {
  my $hmmPos = shift;
  my $hmmFile = shift;

  open(FH,$hmmFile);
  my $start;
  my %hash;
  my ($hmmerPos,$msaPos); 
  while(my $line = <FH>) { 
    last if ($line =~/^\/\//); 
    if ($line=~/^\s+1\s/) { 
      $start = 1; 
    } 

    if ($start) { 
      my ($hmmerPos,$msaPos) = $line=~/^\s+(\d+).*\s+(\d+)$/; 
      if ($hmmerPos == $hmmPos) {
	return $msaPos;
      }
      $line=<FH>; 
      $line=<FH>; 
    } 
  } 
  close(FH); 
}

sub getEvolConsInfo {
  my $sfId = shift;
  my $treeFile = shift;
  my $alignFile = shift;
  my $hmm_col = shift;
  my $wildTypeAA = shift;
  my $mutantAA  = shift;
  my ($pWt,$pSub,$score,$nic,$sfList);
  my $msg;
  my $tmpFile = "$tmpdirPath/consv.$$"; 
  
  my $cmd = "conserved2profile.pl -t $treeFile -p $alignFile -c $hmm_col -w $wildTypeAA -m $mutantAA -compact -T $tmpdirPath "; 
  $cmd .= " -s $sfId" if ($sfId); #if SF ID defined
  $cmd .= " -P $priorFile" if ($priorFile);
  $cmd .= " -S " if ($forceSfFlag);
  $cmd .= " -o $tmpFile";
  `$cmd`;
  my ($sfList,$junk,$evolFam,$sfConsAA);
  if (-s $tmpFile) { 
    #read in outupt file, and get score info
    open(FH,$tmpFile); 
    while(my $line = <FH>) { 
      chomp $line; 
      if ($line=~/^P_wild=(\S+)/) {
	$pWt = $1;
      } elsif ($line=~/^P_sub=(\S+)/) {
	$pSub = $1;
      } elsif ($line=~/^NIC_actual=(\S+)/) {
	$nic = $1;
      } elsif ($line=~/^Conserved SFs=(.*)$/) {
	$sfList = $1;
	$sfList=~s/ /,/g;
	$sfList=~s/,,/,/g;
	$sfList=~s/,$//;
      } elsif ($line=~/Evolutionary family=YES/) {
	$evolFam = "Y";
      } elsif ($line=~/conserved amino acid in SF=(.*)$/) {
	$sfConsAA = $1;
      }
    }
    close(FH); 
  } else { 
    print STDERR "error: could not create $tmpFile from: $cmd\n"; 
    $msg = "profile could not be created";
  }
  unlink($tmpFile);

  unless ($pWt && $pSub && $nic) {
    print STDERR "Missing score info: $pWt - $pSub - $nic!\n";
  }

  my $subPSEC;
  if ($nic == 0) {
      #bug in HMM - this position of the protein ($aaPos) should not align to the HMM (all seqs at position gapped
      $subPSEC = 0;
  } else {
      $subPSEC= &getSubPsec($pWt,$pSub,$nic);
  }

  my $Pdeleterious;
  unless ($subPSEC eq "undef") {
    $Pdeleterious = 1-exp(3+$subPSEC)/(1+exp(3+$subPSEC));
  }
  $Pdeleterious = sprintf "%0.5f",$Pdeleterious;
  $subPSEC = sprintf "%0.5f",$subPSEC if ($subPSEC ne "undef");
  $pWt = sprintf "%0.5f",$pWt if (defined($pWt));
  $pSub = sprintf "%0.5f",$pSub if (defined($pSub));
  $nic = sprintf "%0.5f",$nic if (defined($nic));

  return ($pWt,$pSub,$subPSEC,$Pdeleterious,$nic,$sfList,$evolFam,$sfConsAA,$msg);
} 


sub getSubPsec { 
  my $obsWt = shift; 
  my $obsMut = shift; 
  my $nic = shift; 
 
  #now calculate min/max ratio - take minimum of WT or MUT; divide by MAX  
  #in other words, try to get the ratio small  
 
  my $subPSEC; 
 
  if (! ($obsWt && $obsMut)) {  #zero or undef  
    print STDERR "Missing a probability $obsWt - $obsMut!\n"; 
    $subPSEC = "undef"; 
  } else { 
    my ($min,$max); 
    if ($obsWt < $obsMut) {  
      $min = $obsWt; 
      $max = $obsMut; 
    } else {  
      $min = $obsMut; 
      $max = $obsWt; 
    } 
    $subPSEC = 0.89*log($min) - 0.88*log($max) - 0.94*log($nic); 

  }  
 
  return ($subPSEC);  
} 


sub isSnpPositionValid {
  my $pos = shift;
  my $seq = shift;

  my ($seqText,$title) = FastaFile::splitFasta($seq);
  $seqText=~s/\s+//g;
  my $length = length($seqText);
  if ( ($pos < 1) || ($pos > $length)) {
    return undef;
  } else {
    return 1;
  }
}

sub extractHmmerHmmPos {
  my $seq = shift;
  my $pos = shift;
  my $hmmFile = shift;
  
  my $tmp = "$tmpdirPath/snpanalysis.$$.fasta";
  open(FH,">$tmp");
  print FH $seq;
  close(FH);

  my $args = {}; 
  $args->{'hmmsearchBinary'} = $hmmsearchPath if ($hmmsearchPath); 
  my $hmmerObj = run Hmmer($tmp,$hmmFile,$args); 
  my $hit = $hmmerObj->nextHit();
  my $hmmPos = $hit->proteinPosToHmmPos($pos);
  unlink $tmp;
  return $hmmPos;
}



####
# sub createOutput - creates output line
####
sub createOutput {
  my $output = shift;
  my $aa1 = shift;
  my $aa2 = shift;

  if ($aa1 && $aa2) {
    $output->{"wtAA"} = $aa1;
    $output->{"subAA"} = $aa2;
  }

  my $str;
  foreach my $header (@headerArray) {
    $str .= $output->{$header} . "\t";
  }
  $str=~s/\t$//;
  return $str;
}

####
# sub storeBlosum - stores blosum62 matrix into hash
####
sub storeBlosum {
  my $blosumFile = shift;
  my %matrixHash;
  my $firstLineReached;
  my %posOfAAs;
  open(BLOSUM,$blosumFile) || die "Cannot access blosum 62 matrix\n";
  while (my $line = <BLOSUM>) {
    next if ($line=~/\#/);
    my @array = split(/\s+/,$line);
    
    if (! $firstLineReached) {
      $firstLineReached = 1;
      #first line contains amino acid an position in matrix
      my $colNum;
      foreach my $aa (@array) {
	next unless ($aa=~/\S+/);
	$colNum++;
	$posOfAAs{$colNum} = $aa;
      }
      next;
    }
    
    #if not line 1, store info
    my $rowAA = $array[0];
    my $firstColReached;
    my $colNum;
    foreach my $score (@array) {
      if (! $firstColReached) {
	$firstColReached = 1;
	next;
      }
      $colNum++;
      my $currentAA = $posOfAAs{$colNum};
      $matrixHash{$rowAA}{$currentAA} = $score;
    } 
  }
  close(BLOSUM);
  return(%matrixHash);
}

####
# sub readFastaFile - reads in fasta file and stores id->seq into 
#                     hash for easy, quick access later
####
sub readFastaFile {
  my $fastaFile = shift;
  my %idToSeq;

  my $ff = new FastaFile($fastaFile);
  $ff->open();
  while(my $seq = $ff->nextSeq()) {
    chomp $seq;
    my $id = FastaFile::id($seq);
    $id=~s/\.\d+$//;
    $idToSeq{$id} = $seq;
  }
  $ff->close();

  return(\%idToSeq);
}



####
# sub checkWildType checks that wild type amino acid is actually correct;
#       if not, will try to switch wildtype and mutant
####
sub getWildTypeAA {
  my $gi = shift;
  my $pos = shift;
  my $seq = shift;
  
  my ($seqText,$title) = FastaFile::splitFasta($seq);
  $seqText =~s/\s//g;
  my @aminoAcids = split(//,$seqText);
  my $actualAA = $aminoAcids[$pos - 1];
  return $actualAA;
}





####
# sub getBestProtForSnp - for each SNP, finds best protein (based on hmm scr)
####
sub getBestProtForSnp {
  my $hmmScores = shift;
  my $snpFile = shift;
  my %bestProtForSnp;
  my %snpToProt;
  open(SNP,$snpFile);
  while(my $line = <SNP>) {
    my ($snp,$gi) = split(/\|/,$line);
    $snpToProt{$snp}{$gi}=1;
  }
  close(SNP);

  foreach my $snp (keys %snpToProt) {
    my ($bestProt,$bestScore);
    foreach my $gi (keys %{$snpToProt{$snp}}) {
      my $hmmScore = $hmmScores->{$gi}{'score'};
      if ( (! $bestProt) || ($hmmScore < $bestScore)) {
	$bestProt = $gi;
	$bestScore = $hmmScore;
      }
    }
    $bestProtForSnp{$snp}=$bestProt;
  }
  return (\%bestProtForSnp);
}

#####
# sub readClassFile - reads in class file, and returns best ghmm and shmm
#                     (with scores) for each seq in the file
#####
sub readClassFile {
    my $classFile = shift;
    my $idToSeq = shift;

    my %hmmScores;

    open(CLASS,"$classFile");
    while(my $line=<CLASS>) {
	chomp $line;
	next unless ($line=~/\w/);
	my ($seqId,$hmmId,$hmmName,$score) = split(/\t/,$line);
	$seqId=~s/\.\d+$//;

	unless (defined($idToSeq->{$seqId})) {
	    print STDERR "Warning - class.ts file contains $seqId - not exist in fasta\n";
	}
	if (defined($hmmScores{$seqId})) {
	    print STDERR "Error - died - $seqId appears multiple times in classification file\n";
	    die;
	}

	$hmmScores{$seqId}{'id'} = $hmmId;
	$hmmScores{$seqId}{'score'} = $score;
	$hmmScores{$seqId}{'name'} = $hmmName;
    }
    close(CLASS);
    return(\%hmmScores);
}

sub pukeSetup {
    print STDERR "_"x50, "\n";
    print STDERR "Verbose level is high.\n";
    print STDERR "Output file is: $outFile\n";
    print STDERR "Error file is: $errFile\n";
    print STDERR "Library is: $library\n";
    print STDERR "Entry type is: $type\n" unless ($prod);
    print STDERR "Production version of library being used\n" if ($prod);
    print STDERR "SNP input file is: $snpFile\n";
    print STDERR "fasta file is: $fastaFile\n";
    print STDERR "class.ts file is: $classFile\n";
    print STDERR "NOT RECOMMENDED: -S option used; evolSF will not be used\n" if ($forceSfFlag);
    print STDERR "_"x50, "\n";
}

sub usage {
    my $error = shift;

    print "Error: $error\n\n" if ($error);

    print <<__EOT;

snp_analysis.pl - takes input SNP file generated by the east coast and 
                  Panther scores of corresponding sequences, and provides
		  statistics for the SNP by providing statics for probability
		  of a substitution in the corresponding position of the HMM

Usage:
snp_analysis.pl -l library -c classTsFile -s SNP_file 
                -f fastaFile -V -o output -t entryType

Where args are:
\t-h for help (this message)
\t-l (l)ibrary
\t-t entry (t)ype
\t\tdefault: orig
\t-T for path to  temp directory to store files
\t\tdefault: /tmp
\t-c (c)lassification file contain sequence to PTHR mapping
\t-s (s)np file
\t\tformat: snpId\\tseqID\\tpos\\taminoAcidList
\t\tformat: amino acid list semi-colon separated
\t-f (f)asta file of sequences to be analyzed
\t-P library is in (P)roduction version structure
\t\tdefault: -P 1 (production version of library)
\t\tuse -P 0 for development version of library
\t-a output (a)ll snps, not just best info of best
\t\tprotein for each snp
\t-b for (b)losum62 matrix file
\t-p for (p)rior file
\t-H for path to hmmsearch binary
\t-S for force snp analysis to be on (S)F and not use evol SF (not recommended)
\t-o (o)utput file (redirect STDOUT)
\t-e (e)rror file (redirect STDERR)
\t-v (v)erbose (debug info to STDERR)
\t-V (V)ery verbose (debug info to STDERR)
__EOT

exit(-1);
}
